import sys
import numpy as np
from numpy.typing import ArrayLike, NDArray
from typing import Union


def check_single_array(
        a: ArrayLike,
        *,
        is_clean: bool = False,
        is_flatten: bool = False,
        is_nan_warning: bool = False,
        nan_warning_criterion: float = 0.3
) -> NDArray:
    """
        检查并转换两个数组
        :param a: ArrayLike
        :param is_clean: 是否去除 nan 值
        :param is_flatten: 是否展平
        :param is_nan_warning: nan 值过多时是否提出 warning
        :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
        :return:
        """
    a = np.asarray(a)

    if is_clean:
        a = a[~np.isnan(a)]

    if is_flatten:
        a = a.flatten()

    if is_nan_warning:
        _nan_warning(a, nan_warning_criterion=nan_warning_criterion)

    return a


def check_two_array(
        a: ArrayLike, b: ArrayLike,
        *,
        is_same_size: bool = False,
        is_same_shape: bool = False,
        is_clean_each: bool = False,
        is_clean_same_time: bool = False,
        is_flatten: bool = False,
        is_nan_warning: bool = False,
        nan_warning_criterion: float = 0.3
) -> (NDArray, NDArray):
    """
    检查并转换两个数组
    :param a: ArrayLike
    :param b: ArrayLike
    :param is_same_size: 是否检查相同 size
    :param is_same_shape: 是否检查相同 shape
    :param is_clean_each: 是否分别清理 nan 值
    :param is_clean_same_time: 是否由两个数组的 nan 值共同决定清洗
    :param is_flatten: 是否展平
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return:
    """
    a = np.asarray(a)
    b = np.asarray(b)

    if is_same_size:
        if a.size != b.size:
            raise ValueError(f'Expect two array have then same size, but got {a.size} and {b.size}')

    if is_same_shape:
        if a.shape != b.shape:
            raise ValueError(f'Expect two array have then same shape, but got {a.shape} and {b.shape}')

    if is_clean_each:
        a = a[~np.isnan(a)]
        b = b[~np.isnan(b)]
    if is_clean_same_time:
        not_nan = np.logical_not(np.logical_or(np.isnan(a), np.isnan(b)))
        a = a[not_nan]
        b = b[not_nan]

    if is_flatten:
        a = a.flatten()
        b = b.flatten()

    if is_nan_warning:
        _nan_warning(a, b, nan_warning_criterion=nan_warning_criterion)

    return a, b


def check_three_array(
        a: ArrayLike, b: ArrayLike, c: ArrayLike,
        *,
        is_same_size: bool = False,
        is_same_shape: bool = False,
        is_clean_each: bool = False,
        is_clean_same_time: bool = False,
        is_flatten: bool = False,
        is_nan_warning: bool = False,
        nan_warning_criterion: float = 0.3
):
    """
    检查并转换三个数组
    :param a: ArrayLike
    :param b: ArrayLike
    :param c: ArrayLike
    :param is_same_size: 是否检查相同 size
    :param is_same_shape: 是否检查相同 shape
    :param is_clean_each: 是否分别清理 nan 值
    :param is_clean_same_time: 是否由两个数组的 nan 值共同决定清洗
    :param is_flatten: 是否展平
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return:
    """
    a = np.asarray(a)
    b = np.asarray(b)
    c = np.asarray(c)

    if is_same_size:
        if a.size != b.size != b.size:
            raise ValueError(f'Expect two array have then same size, but got {a.size} and {b.size}')

    if is_same_shape:
        if a.shape != b.shape != c.shape:
            raise ValueError(f'Expect two array have then same shape, but got {a.shape} and {b.shape}')

    if is_clean_each:
        a = a[~np.isnan(a)]
        b = b[~np.isnan(b)]
        c = c[~np.isnan(c)]
    if is_clean_same_time:
        _t = np.logical_or(np.isnan(a), np.isnan(b))
        not_nan = np.logical_not(np.logical_or(_t, np.isnan(c)))
        a = a[not_nan]
        b = b[not_nan]
        c = c[not_nan]

    if is_flatten:
        a = a.flatten()
        b = b.flatten()
        c = c.flatten()

    if is_nan_warning:
        _nan_warning(a, b, c, nan_warning_criterion=nan_warning_criterion)

    return a, b, c


def _nan_warning(*arrays: ArrayLike, nan_warning_criterion: float):
    """
    （内部方法）封装 nan 值校验
    :param arrays:
    :param nan_warning_criterion:
    :return:
    """
    nan_warning_criterion = check_float(
        nan_warning_criterion,
        is_lower_closed=True,
        is_upper_closed=True,
        minimum=0, maximum=1
    )
    for array in arrays:
        if np.count_nonzero(np.isnan(array)) / array.size >= array.size * nan_warning_criterion:
            raise RuntimeWarning('Input array nan value have too much')


def check_int(
        a: Union[int, float],
        *,
        is_upper_closed: bool = False,
        is_lower_closed: bool = False,
        minimum: int = -sys.maxsize,
        maximum: int = sys.maxsize
) -> int:
    """
    检查整数是否在范围中
    :param a: 整数
    :param is_upper_closed: 上限是否为闭区间
    :param is_lower_closed: 下限是否为闭区间
    :param minimum: 区间最小值
    :param maximum: 区间最大值
    :return:
    """
    a = int(a)
    if not minimum <= a <= maximum:
        raise ValueError
    if not is_lower_closed and a == minimum:
        raise ValueError
    if not is_upper_closed and a == maximum:
        raise ValueError

    return a


def check_float(
        a: Union[int, float],
        *,
        is_upper_closed: bool = False,
        is_lower_closed: bool = False,
        minimum: int = -sys.maxsize,
        maximum: int = sys.maxsize
) -> float:
    """
    检查整数是否在范围中
    :param a: 整数
    :param is_upper_closed: 上限是否为闭区间
    :param is_lower_closed: 下限是否为闭区间
    :param minimum: 区间最小值
    :param maximum: 区间最大值
    :return:
    """
    a = float(a)
    if not minimum <= a <= maximum:
        raise ValueError
    if not is_lower_closed and a == minimum:
        raise ValueError
    if not is_upper_closed and a == maximum:
        raise ValueError

    return a


def rolling(a: ArrayLike, win: int):
    """
    滑动窗口
    :param a:
    :param win:
    :return:
    """
    a = np.asarray(a)
    win = check_int(win, minimum=1, maximum=np.min(a.shape), is_lower_closed=True, is_upper_closed=True)
    shape = (a.size - win + 1, win)
    strides = (a.itemsize, a.itemsize)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)



from enum import Enum
from numpy import arcsin, cos, radians, sin, sqrt


class DistanceUnit(Enum):
    """
    距离单位
    """
    KILOMETERS = 'km'
    METERS = 'm'
    MILES = 'mi'


# 平均地球半径 - https://en.wikipedia.org/wiki/Earth_radius#Mean_radius
AVG_EARTH_RADIUS_M = 6371008


# 距离换算因数
_CONVERSIONS_FROM_M = {
    DistanceUnit.KILOMETERS: 1000.0,
    DistanceUnit.METERS: 1.0,
    DistanceUnit.MILES: 0.000621371192
}


def distance(point1, point2, *, unit: DistanceUnit = DistanceUnit.KILOMETERS) -> float:
    """
    计算大圆距离
    :param point1: 经纬度坐标点，形式为 (latitude, longitude)
    :param point2: 经纬度坐标点，形式为 (latitude, longitude)
    :param unit: 距离单位，默认为米
    :return: 大圆距离
    """
    # 解算经纬度
    lat1, lng1 = point1
    lat2, lng2 = point2

    # 转为弧度制
    lat1 = radians(lat1)
    lng1 = radians(lng1)
    lat2 = radians(lat2)
    lng2 = radians(lng2)

    # 计算
    lat = lat2 - lat1
    lng = lng2 - lng1
    d = sin(lat * 0.5) ** 2 + cos(lat1) * cos(lat2) * sin(lng * 0.5) ** 2

    return 2 * (AVG_EARTH_RADIUS_M * _CONVERSIONS_FROM_M[DistanceUnit(unit)]) * arcsin(sqrt(d))



from enum import Enum
try:
    from shapely.geometry import shape
except ModuleNotFoundError:
    raise ModuleNotFoundError('Expect package named shapely, try to install shapely')
from pyproj import CRS, Proj


class AreaUnit(Enum):
    """
    面积单位
    """
    KILOMETERS2 = 'km2'
    METERS2 = 'm2'


# 面积换算因数
_CONVERSIONS_FROM_M2 = {
    AreaUnit.KILOMETERS2: 1e+6,
    AreaUnit.METERS2: 1.0
}


def area(*points, unit: AreaUnit = AreaUnit.METERS2, crs: CRS = CRS.from_epsg(3857)) -> float:
    """
    计算地球上多边形的面积
    :param points: 多个点集合，形式为 (latitude, longitude)
    :param unit: 面积单位，默认为平方米
    :param crs: CRS 实例，默认为墨卡托投影
    :return: 面积
    """
    p = Proj(crs.to_proj4())

    return shape({
        'type': 'Polygon',
        'coordinates': [list(map(lambda x: p(*x[:: -1]), points))]
    }).area * _CONVERSIONS_FROM_M2[unit]



import numpy as np
import scipy as sp
import scipy.stats
from numpy.typing import ArrayLike
from .tool import check_two_array, check_single_array, check_int, rolling
from typing import NamedTuple
from collections import namedtuple


def difference_test(a: ArrayLike, b: ArrayLike) -> NamedTuple:
    """
    样本差异性检验，原假设为方差相等
    :param a: 样本 1
    :param b: 样本 2
    :return: VarianceDifference(coefficient, pvalue)
    """
    a, b = check_two_array(a, b, is_clean_each=True, is_flatten=True)

    f = a.var() / b.var()

    return namedtuple(
        'VarianceDifferenceResult',
        ('statistic', 'pvalue')
    )(f, 2 * sp.stats.f.sf(f, a.size - 1, b.size - 1))


def auto_cov(a: ArrayLike, tau: int = 0, bias: bool = False) -> float:
    """
    自协方差
    :param a: ArrayLike
    :param tau: 滞后步长
    :param bias:
        Default normalization (False) is by ``(N - 1)``, where ``N`` is the
        number of observations given (unbiased estimate). If `bias` is True,
        then normalization is by ``N``.
    :return:
    """
    a = check_single_array(a, is_clean=True, is_flatten=True)
    tau = check_int(tau, minimum=0, maximum=a.size - 1, is_lower_closed=True, is_upper_closed=True)
    s = rolling(a, tau + 1)

    return np.sum((s[:, 0] - a.mean()) * (s[:, -1] - a.mean())) / (a.size if bias else a.size - 1)


def cross_cov(a: ArrayLike, b: ArrayLike, tau: int = 0, bias: bool = False) -> float:
    """
    交叉协方差
    :param a: ArrayLike
    :param b: ArrayLike
    :param tau: 滞后步长
    :param bias:
        Default normalization (False) is by ``(N - 1)``, where ``N`` is the
        number of observations given (unbiased estimate). If `bias` is True,
        then normalization is by ``N``.
    :return:
    """
    a, b = check_two_array(
        a, b, is_same_size=True, is_same_shape=True,
        is_clean_same_time=True, is_flatten=True
    )
    tau = check_int(tau, minimum=0, maximum=a.size - 1, is_lower_closed=True, is_upper_closed=True)
    sa = rolling(a, tau + 1)
    sb = rolling(b, tau + 1)

    return np.sum((sa[:, 0] - a.mean()) * (sb[:, -1] - b.mean())) / (a.size if bias else a.size - 1)



import numpy as np
import scipy as sp
import scipy.stats
from numpy.typing import ArrayLike
from .tool import check_single_array, check_three_array, check_int, rolling
from typing import NamedTuple
from collections import namedtuple


def partial3var(a: ArrayLike, b: ArrayLike, c: ArrayLike) -> NamedTuple:
    """
    三变量偏相关系数
    :param a:
    :param b:
    :param c:
    :return:
    """
    a, b, c = check_three_array(
        a, b, c,
        is_same_size=True,
        is_same_shape=True,
        is_clean_same_time=True,
        is_flatten=True
    )

    p_ab, _ = sp.stats.pearsonr(a, b)
    p_ac, _ = sp.stats.pearsonr(a, c)
    p_bc, _ = sp.stats.pearsonr(b, c)

    stat = (p_ab - p_ac * p_bc) / np.sqrt((1 - p_ac**2) * (1 - p_bc**2))

    t = stat * np.sqrt(a.size - 5) / np.sqrt(1 - stat**2)

    return namedtuple(
        'PartialThreeVariableResult',
        ('statistic', 'pvalue')
    )(t, sp.stats.t.sf(t, a.size - 5))



import base64
import json
import zlib


def encode(origin):
    compressed = zlib.compress(json.dumps(origin).encode())

    return base64.b64encode(compressed)[:: -1].decode()


def decode(origin):
    decompressed = base64.b64decode(origin[:: -1])

    return json.loads(zlib.decompress(decompressed))



from setuptools import find_packages, setup


VERSION = '0.0.3'

setup(
    name='queensbarry',
    version=VERSION,
    description='Self use tool.',
    packages=find_packages(exclude=('tests*',)),
    author='Queensbarry',
    author_email='queensbarry@foxmail.com',
    url='https://github.com/Queensbarry/qb',
    license='GPL License',
    python_requires='>3.8',
    include_package_data=True,
    install_requires=[
        'numpy>=1.21.0'
    ],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ]
)



import numpy as np
from numpy.typing import ArrayLike
from .tool import check_two_array
from typing import Callable


def mean_error(a: ArrayLike, b: ArrayLike, *, is_nan_warning=True, nan_warning_criterion=0.3) -> float:
    """
    平均误差
    :param a: ArrayLike
    :param b: ArrayLike
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return: 平均误差值
    """
    a, b = check_two_array(
        a, b,
        is_same_size=True,
        is_same_shape=True,
        is_clean_same_time=True,
        is_nan_warning=is_nan_warning,
        nan_warning_criterion=nan_warning_criterion
    )

    return (a - b).sum() / a.size


def mean_absolute_error(a: ArrayLike, b: ArrayLike, *, is_nan_warning=True, nan_warning_criterion=0.3):
    """
    平均绝对误差
    :param a: ArrayLike
    :param b: ArrayLike
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return: 平均绝对误差值
    """
    a, b = check_two_array(
        a, b,
        is_same_size=True,
        is_same_shape=True,
        is_clean_same_time=True,
        is_nan_warning=is_nan_warning,
        nan_warning_criterion=nan_warning_criterion
    )

    return np.abs(a - b).sum() / a.size


def relative_absolute_error(a: ArrayLike, b: ArrayLike, *, is_nan_warning=True, nan_warning_criterion=0.3):
    """
    相对绝对误差
    :param a: ArrayLike
    :param b: ArrayLike
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return: 相对绝对误差值
    """
    a, b = check_two_array(
        a, b,
        is_same_size=True,
        is_same_shape=True,
        is_clean_same_time=True,
        is_nan_warning=is_nan_warning,
        nan_warning_criterion=nan_warning_criterion
    )

    return np.abs((a - b) / b).sum() / a.size


def root_mean_squared_error(a: ArrayLike, b: ArrayLike, *, is_nan_warning=True, nan_warning_criterion=0.3):
    """
    均方根误差
    :param a: ArrayLike
    :param b: ArrayLike
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return: 均方根误差值
    """
    a, b = check_two_array(
        a, b,
        is_same_size=True,
        is_same_shape=True,
        is_clean_same_time=True,
        is_nan_warning=is_nan_warning,
        nan_warning_criterion=nan_warning_criterion
    )

    return np.sqrt(np.power(a - b, 2).sum() / (a - b).sum())


def mean_squared_error(a: ArrayLike, b: ArrayLike, *, is_nan_warning=True, nan_warning_criterion=0.3):
    """
    均方误差
    :param a: ArrayLike
    :param b: ArrayLike
    :param is_nan_warning: nan 值过多时是否提出 warning
    :param nan_warning_criterion: nan 值超过 size 的百分率 warning（仅在 is_nan_warning=True 时生效）
    :return: 均方误差差值
    """
    a, b = check_two_array(
        a, b,
        is_same_size=True,
        is_same_shape=True,
        is_clean_same_time=True,
        is_nan_warning=is_nan_warning,
        nan_warning_criterion=nan_warning_criterion
    )

    return np.power(a - b, 2).sum() / (a - b).sum()


# --- 以下为别名 ---
me: Callable = mean_error
mae: Callable = mean_absolute_error
rae: Callable = relative_absolute_error
rmse: Callable = root_mean_squared_error
mse: Callable = mean_squared_error
